"""Functions for analysis."""
import re


def _find_cross_feed(row, tol=1e-4):
    positive = set()
    negative = set()
    for rid, val in row.items():
        compound = re.search("(?<=R_EX_).*(?=_e)", rid)
        if compound and rid.endswith("_i"):
            global_rid = f"R_EX_{compound}_e"
            global_val = row[global_rid] if global_rid in row else 0
            adjusted_val = val - global_val
            if adjusted_val < -tol:
                negative.add(compound.group(0))
            elif adjusted_val > tol:
                positive.add(compound.group(0))
    return positive & negative


def _find_directed_cross_feed(row, tol=1e-4):
    strain = row.name[1]
    crossfeed = _find_cross_feed(row, tol)
    directed_crossfeed = {}
    for compound in crossfeed:
        rid = f"R_EX_{compound}_e_{strain}_i"
        directed_crossfeed[compound] = row[rid] if rid in row else 0
    return directed_crossfeed


def compute_crossfeed(data_frame, tol=1e-4):
    return data_frame.apply(_find_cross_feed, axis=1, tol=tol)


def compute_directed_crossfeed(data_frame, tol=1e-4):
    return data_frame.apply(_find_directed_cross_feed, axis=1, tol=tol)


def find_suppliers(data_frame):
    """Find suppliers.

    Args:
        data_frame (DataFrame): MiSoS(oup) solution.
    """

    def _find_suppliers(row):
        suppliers = set()
        for col, value in row.items():
            if col.startswith("y_") and value > 0.5:
                suppliers.add(col)
        return suppliers

    data_frame["suppliers"] = data_frame.apply(_find_suppliers, axis=1)
    return data_frame


def find_focal_strain_growth(data_frame):
    """Find focal strain growth and add to DataFrame.
    Args:
        data_frame (DataFrame): MiSoS(oup) solution.
    """
    data_frame["strain_growth"] = data_frame.apply(
        lambda row: row[f"Growth_{row.name[1]}"], axis=1
    )
    return data_frame



"""MiSoS(oup)"""

import argparse
import glob
import logging
import os
from collections import defaultdict

import yaml
import gurobipy as gp

from .library.getters import get_biomass, get_exchange_reactions
from .library.load import introduce_binary_variables, setup_medium
from .library.readwrite import load_models, read_compounds
from .library.solve import minimal_communities, minimal_suppliers
from .library.validate import validate_solution_dict

from .reframed.layered_community import LayeredCommunity
from .reframed.gurobi_env_solver import GurobiEnvSolver

compute_function = defaultdict(lambda: minimal_suppliers, {"min": minimal_communities})


def compute_solution(
    org_id,
    medium,
    community,
    objective,
    parsimony,
    community_size,
    minimal_growth,
):
    """Setup and execute misosoup."""
    with gp.Env(params={"Method": 1, "LogToConsole": 0}) as env:
        solver = GurobiEnvSolver(community.merged_model, env)

        introduce_binary_variables(community, solver, minimal_growth=minimal_growth)

        setup_medium(community.merged_model, solver, medium)

        solution = compute_function[org_id](
            org_id,
            community,
            solver,
            values=(
                get_biomass(community) + get_exchange_reactions(community.merged_model)
            ),
            community_size=community_size,
            objective=objective,
            parsimony=parsimony,
            minimal_growth=minimal_growth,
        )

    return (org_id, solution)


def main(args):
    """Main function."""
    input_paths = glob.glob(args.input[0]) if len(args.input) == 1 else args.input
    models = load_models(input_paths)

    media = read_compounds(args.media)
    base_medium = media["base_medium"] if "base_medium" in media.keys() else {}

    community = LayeredCommunity("CoI", models, copy_models=False)

    solution = defaultdict(dict)

    if args.objective:
        objective = {reaction: 1 for reaction in args.objective}
    else:
        objective = {community.merged_model.biomass_reaction: 1}

    if args.disable_objective:
        objective = None

    for medium_id, medium_composition in media.items():
        if not medium_id == "base_medium":
            medium = {
                **medium_composition,
                **base_medium,
            }
            org_id, sol = compute_solution(
                args.strain,
                medium=medium,
                community=community,
                community_size=args.community_size,
                objective=objective,
                parsimony=args.parsimony,
                minimal_growth=args.minimal_growth,
            )

            solution[medium_id][org_id] = sol

    output_dict = {
        k: {
            org: [{var: rate for var, rate in s.values.items() if rate} for s in sol]
            if sol
            else [{f"Growth_{org}": 0}]
            for org, sol in v.items()
        }
        for k, v in solution.items()
    }

    if args.validate:
        validate_solution_dict(output_dict, args.exchange_format)

    output = yaml.dump(output_dict)

    if args.output:
        if not os.path.exists(os.path.dirname(args.output)):
            os.makedirs(os.path.dirname(args.output))
        with open(args.output, "w", encoding="utf8") as file_descriptor:
            file_descriptor.write(output)
    else:
        print(output)


def entry():
    """Misosoup entry point."""
    parser = argparse.ArgumentParser(
        description="""
    Compute minimal supplying communities with MiSoS(oup).

    `misosoup` will create a large community network and evaluate which
    communities can supply growth of some strain (or each other, if no strain
    is chosen).

    It will report the results in `yaml` format, where the dictionaries contain
    all active exchange reactions and their rates, as well as the activity of
    strains (i.e. `y_strain: 1` for any active strain), the community growth
    rate and the individual growth rates of all participating strains:

        ```
        carbon_source:
            strain:
                - {}
                - {}
        ```
""",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "input", nargs="+", type=str, help="List or wildcard for model paths."
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="Output file. Format: YAML. If not supplied, will print to stdout.",
    )
    parser.add_argument(
        "--media",
        type=str,
        required=True,
        help="Path to media. Format: YAML. File needs to contain dictionary with media, see examples.",
    )
    parser.add_argument(
        "--strain",
        type=str,
        default="min",
        help="Focal strain model id. If no strain is provided, we compute minimal communities.",
    )
    parser.add_argument(
        "--parsimony", action="store_true", help="Compute parsimony solution."
    )
    parser.add_argument(
        "--community-size",
        type=int,
        default=0,
        help="Maximum community size. Default: 0 (Arbitrary).",
    )
    parser.add_argument(
        "--minimal-growth",
        type=float,
        default=0.01,
        help=(
            "Minimal required growth for strain or community."
            "Each strain that is considered to grow needs to at least achieve "
            "this minimal growth rate given by this argument."
        ),
    )
    parser.add_argument(
        "--exchange-format",
        type=str,
        default="R_EX_{}_e",
        help=(
            "Regular expression to retrieve the carbon source from an exchange "
            "reaction. The group containing the carbon source, should be named "
            "`carbon_source`; as can be seen from the default. "
            "default: `R_EX_(\\w+)_e"
        ),
    )
    parser.add_argument(
        "--disable-objective",
        action="store_true",
        help="If set, no objective will be optimized.",
    )
    parser.add_argument(
        "--objective",
        type=str,
        nargs="+",
        help=(
            "List of ids that are part of the objective function. By default "
            "the community biomass is maximized."
        ),
    )
    parser.add_argument(
        "--validate",
        action="store_true",
        help="Validate solution. Numerical consistency will verified.",
    )
    parser.add_argument("--log", default="INFO", help="Log level. Default: INFO")

    args_parsed = parser.parse_args()

    logging.basicConfig(level=args_parsed.log, format="%(asctime)s %(message)s")

    main(args_parsed)



"""Solver functions."""
import logging

from reframed.solvers.solution import Status

logger = logging.getLogger(__name__)


def minimal_suppliers(
    org,
    community,
    solver,
    values,
    community_size=0,
    objective=None,
    parsimony=False,
    **kwargs,
):
    """Compute minimal suppliers with MiSoS(oup).

    Keyword Arguments:
    org -- focal strain id
    community -- community model
    solver -- solver instance
    values -- list with variables to store
    community_size -- max community size, if 0 all community sizes will be searched (default: 0)
    growth -- optimize growth (default: True)
    parsimony -- minimize flux (default: False)
    """
    solver.add_constraint(
        f"c_{org}_focal",
        {f"y_{org}": 1},
        ">",
        1,
        update=True,
    )
    solutions = _minimize(
        community, solver, values, community_size, objective, parsimony
    )
    solver.remove_constraints([f"c_{org}_growth"])
    return solutions


def minimal_communities(
    org_id,
    community,
    solver,
    values,
    community_size=0,
    objective=None,
    parsimony=False,
    minimal_growth=0.01,
):
    """Compute minimal communities with MiSoS(oup).

    Keyword Arguments:
    community -- community model
    solver -- solver instance
    values -- list with variables to store
    community_size -- max community size, if 0 all community sizes will be searched (default: 0)
    growth -- optimize growth (default: True)
    parsimony -- minimize flux (default: False)
    """
    assert org_id == "min"
    solver.add_constraint(
        "c_community_growth",
        {community.merged_model.biomass_reaction: 1},
        ">",
        minimal_growth,
        update=True,
    )
    solutions = _minimize(
        community, solver, values, community_size, objective, parsimony
    )
    solver.remove_constraints(["c_community_growth"])
    return solutions


def _minimize(community, solver, values, community_size, objective, parsimony):
    """Minimize the community size for problem as set up in solver."""
    solutions = []
    constraints = []
    knowledge_constraints = []

    obj = {f"y_{org_id}": 1 for org_id in community.organisms.keys()}

    solution_length = 0
    i = 0

    while True:
        logging.info("------------")
        logging.info("Starting optimization...")
        solution = solver.solve(
            linear=obj,
            get_values=list(obj.keys()) + values,
            minimize=True,
        )

        logging.info("Solution status: %s", str(solution.status))

        if solution.status != Status.OPTIMAL:
            break

        selected = {k: 1 for k in obj.keys() if solution.values[k] > 0.5}
        not_selected = {k: 1 for k in obj.keys() if solution.values[k] < 0.5}
        logging.info("Community size: %i", len(selected))
        logging.info(
            "Community growth: %f",
            solution.values[community.merged_model.biomass_reaction],
        )

        # Fix selected community
        solver.add_constraint("c_not_selection", not_selected, "=", 0, update=True)
        solver.add_constraint("c_selection", selected, "=", len(selected), update=True)

        try:
            if objective:
                logging.info("Starting growth optimization.")
                solution = solver.solve(
                    linear=objective,
                    get_values=list(obj.keys()) + values,
                    minimize=False,
                )
                if solution.status != Status.OPTIMAL:
                    logging.info(
                        "Community Inconsistent: %s", str(list(selected.keys()))
                    )
                    solver.add_constraint(
                        f"c_tmp_{len(knowledge_constraints) + 1}", not_selected, ">", 1
                    )
                    knowledge_constraints.append(
                        f"c_tmp_{len(knowledge_constraints) + 1}"
                    )
                    continue
                logging.info(
                    "Community growth: %f",
                    solution.values[community.merged_model.biomass_reaction],
                )

            if parsimony:
                logging.info("Starting parsimony optimization.")
                if objective:
                    solver.add_constraint(
                        "c_growth",
                        objective,
                        ">",
                        solution.values[community.merged_model.biomass_reaction] - 1e-4,
                        update=True,
                    )

                solution = solver.solve(
                    quadratic={
                        (rid, rid): 1
                        for rid in community.merged_model.reactions
                        if not rid.startswith("Growth")
                        and rid != community.merged_model.biomass_reaction
                    },
                    get_values=list(obj.keys()) + values,
                    minimize=True,
                )

                solver.remove_constraints(["c_growth"])

                logging.info("Parsimony solution status: %s", str(solution.status))
                if solution.status != Status.OPTIMAL:
                    logging.info(
                        "Community Inconsistent: %s", str(list(selected.keys()))
                    )
                    solver.add_constraint(
                        f"c_tmp_{len(knowledge_constraints) + 1}", not_selected, ">", 1
                    )
                    knowledge_constraints.append(
                        f"c_tmp_{len(knowledge_constraints) + 1}"
                    )
                    continue
                logging.info(
                    "Parsimony community growth: %f",
                    solution.values[community.merged_model.biomass_reaction],
                )

            if not (objective or parsimony):
                logging.info("Starting no optimization.")
                solver.solve(get_values=values)
                if solution.status != Status.OPTIMAL:
                    logging.info(
                        "Community Inconsistent: %s", str(list(selected.keys()))
                    )
                    solver.add_constraint(
                        f"c_tmp_{len(knowledge_constraints) + 1}", not_selected, ">", 1
                    )
                    knowledge_constraints.append(
                        f"c_tmp_{len(knowledge_constraints) + 1}"
                    )
                logging.info(
                    "Community growth: %f",
                    solution.values[community.merged_model.biomass_reaction],
                )
        finally:
            solver.remove_constraints(["c_not_selection", "c_selection"])

        solver.add_constraint(f"c_{i}", selected, "<", len(selected) - 1, update=True)
        constraints.append(f"c_{i}")

        solution_length = len(selected)
        if community_size and solution_length > community_size:
            break

        solutions.append(solution)

        i += 1

    solver.remove_constraints(constraints)
    solver.remove_constraints(knowledge_constraints)

    return solutions



"""MiSoS(oup) utilities."""
import argparse
import sys
import os

from glob import glob
from random import sample

import yaml

from .library.validate import validate_solution_file


def filter_soup():
    """Main."""
    parser = argparse.ArgumentParser(
        description="Remove numerical relicts from `misosoup` solutions."
    )
    parser.add_argument("--log", default="INFO")
    parser.add_argument("input", type=str)
    parser.add_argument("-o", "--output", type=str)
    parser.add_argument("--threshold", type=float, default=10e-8)

    args = parser.parse_args()

    with open(args.input) as file_descriptor:
        solution_data = yaml.safe_load(file_descriptor)

    output_dict = {
        cs: {
            org: [
                {r: v for r, v in sol.items() if abs(v) > args.threshold}
                for sol in org_sols
            ]
            for org, org_sols in cs_sols.items()
        }
        for cs, cs_sols in solution_data.items()
    }

    output = yaml.dump(output_dict)

    if args.output:
        if not os.path.exists(os.path.dirname(args.output)):
            os.makedirs(os.path.dirname(args.output))
        with open(args.output, "w") as file_descriptor:
            file_descriptor.write(output)
    else:
        print(output)


def taste_soup():
    """Verify integrity of solutions in raw misosoup output."""
    parser = argparse.ArgumentParser(
        description="Verify integrity of solutions in raw `misosoup` output."
    )
    parser.add_argument("--log", default="INFO")
    parser.add_argument("input", type=str)

    args = parser.parse_args()

    if validate_solution_file(args.input):
        print("\033[92mValid solution file.\033[0m")
    else:
        print("\033[91mInvalid solution file.\033[0m")
        sys.exit(os.EX_DATAERR)


def select_ingredients():
    """Create directory with sampled models."""
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("--log", default="INFO")
    parser.add_argument("source", type=str, help="Source directory with sbml models.")
    parser.add_argument(
        "target", type=str, help="Target directory is created and contains symlinks."
    )
    parser.add_argument("-n", "--n-models", type=int, default=10)

    args = parser.parse_args()

    files = glob(os.path.join(args.source, "*.xml"))
    file_names = [os.path.basename(f) for f in files]
    source_list = args.source.split(os.sep)
    target_list = args.target.split(os.sep)

    if not source_list[-1]:
        source_list.pop()
    if not target_list[-1]:
        target_list.pop()

    # move to common node
    while target_list[0] == source_list[0]:
        target_list.pop(0)
        source_list.pop(0)

    # add relative difference
    source_list = len(target_list) * [".."] + source_list

    relative_source = os.path.join(*source_list)
    relative_files = [os.path.join(relative_source, f) for f in file_names]

    os.makedirs(args.target)

    sampled = sample(relative_files, args.n_models)
    for _sample in sampled:
        os.symlink(_sample, os.path.join(args.target, os.path.basename(_sample)))



"""Load and init solvers."""
import logging
import math

from reframed.solvers.solver import VarType


BOUND_INF = 1000


def introduce_binary_variables(community, solver, minimal_growth=0.01):
    """Add binary variables to the solver instance.

    Keyword arguments:
    community -- community model instance
    solver -- solver instance
    min_growth -- minimal growth rate required (default: 0.01)
    """
    for org_id in community.organisms.keys():
        solver.add_variable(f"y_{org_id}", 0, 1, vartype=VarType.BINARY, update=False)

    solver.update()

    for org_id, org_model in community.organisms.items():
        org_var = f"y_{org_id}"
        for r_id, reaction in org_model.reactions.items():
            if (
                not r_id.startswith("R_EX")
                and r_id != org_model.biomass_reaction
                and reaction.lb * reaction.ub <= 0
            ):
                continue

            merged_id = community.reaction_map[(org_id, r_id)]
            ubound = BOUND_INF
            lbound = -BOUND_INF

            if r_id == org_model.biomass_reaction:
                lbound = minimal_growth

            if reaction.lb * reaction.ub > 0:
                lbound = -BOUND_INF if math.isinf(reaction.lb) else reaction.lb
                ubound = BOUND_INF if math.isinf(reaction.ub) else reaction.ub 
                solver.set_bounds({merged_id: (-BOUND_INF, BOUND_INF)})

            solver.add_constraint(
                f"c_{merged_id}_lb",
                {merged_id: 1, org_var: -lbound},
                ">",
                0,
                update=False,
            )
            solver.add_constraint(
                f"c_{merged_id}_ub",
                {merged_id: 1, org_var: -ubound},
                "<",
                0,
                update=False,
            )


def setup_medium(model, solver, medium):
    """Setup the medium for model on solver.

    Arguments:
    model -- model instance
    solver -- solver instance
    medium -- medium dict
    """
    missing_reactions = set(medium.keys()) - set(model.reactions.keys())
    for r_id in missing_reactions:
        logging.warning("Missing reaction %s in model %s", r_id, model.id)
    for r_id in model.reactions.keys():
        if r_id.startswith("R_EX_") and not r_id.endswith("_i"):
            bound = medium[r_id] if r_id in medium.keys() else 0
            solver.add_constraint(
                f"c_{r_id}_lb",
                {r_id: 1},
                ">",
                bound,
            )



"""Common functions."""
import re
import yaml


def get_metabolite_name(compound):
    """Get metabolite id from compound name."""
    return f"M_{compound}_c"


def get_reaction_name(compound):
    """Get reaction id from compound name."""
    return f"EX_{compound}_e"


def get_compound_name(exchange_reaction):
    """Get compound name from reaction id."""
    return re.search(
        "((?<=R_EX_).*(?=_e))|((?<=EX_).*(?=_e))", exchange_reaction
    ).group(0)


def merge_dicts(a, b, path=None):
    """Merge b into a."""
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge_dicts(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            else:
                raise Exception("Conflict at %s" % ".".join(path + [str(key)]))
        else:
            a[key] = b[key]
    return a


def merge_yaml(file_list):
    """Merge yaml files."""
    solutions = {}
    for path in file_list:
        with open(path, "r") as file_descriptor:
            merge_dicts(solutions, yaml.safe_load(file_descriptor))
    return solutions



"""Layered community class implementation."""

from math import inf

from reframed import (
    Community,
    CBModel,
    CBReaction,
    Gene,
    Protein,
    GPRAssociation,
    ReactionType,
    Compartment,
    Metabolite,
)


class LayeredCommunity(Community):
    """Community model with additional layer of exchange reactions for each member."""

    def __init__(self, community_id: str, models: list, copy_models=False, suffix="_i"):
        super().__init__(
            community_id=community_id,
            models=models,
            copy_models=copy_models,
        )
        self.suffix = suffix

    def merge_models(self):
        comm_model = CBModel(self.id)
        old_ext_comps = []
        ext_mets = []
        self.reaction_map = {}
        self.metabolite_map = {}

        # default IDs
        ext_comp_id = "ext"
        biomass_id = "community_biomass"
        comm_growth = "community_growth"

        # create external compartment

        comp = Compartment(ext_comp_id, "extracellular environment", external=True)
        comm_model.add_compartment(comp)

        # community biomass

        met = Metabolite(biomass_id, "Total community biomass", ext_comp_id)
        comm_model.add_metabolite(met)

        rxn = CBReaction(
            comm_growth,
            name="Community growth rate",
            reversible=False,
            stoichiometry={biomass_id: -1},
            lb=0,
            ub=inf,
            objective=1,
        )

        comm_model.add_reaction(rxn)

        # add each organism
        for org_id, model in self.organisms.items():

            def rename(old_id):
                return f"{old_id}_{org_id}"

            # add internal compartments
            for c_id, comp in model.compartments.items():
                if comp.external:
                    old_ext_comps.append(c_id)
                new_comp = Compartment(rename(c_id), comp.name)
                comm_model.add_compartment(new_comp)

            # add metabolites
            for m_id, met in model.metabolites.items():
                # always create a local metabolite
                new_id = rename(m_id)
                new_met = Metabolite(new_id, met.name, rename(met.compartment))
                new_met.metadata = met.metadata.copy()
                comm_model.add_metabolite(new_met)
                self.metabolite_map[(org_id, m_id)] = new_id

                # if the metabolite is external and not yet added to the
                # community model, add it
                if (
                    met.compartment in old_ext_comps
                    and m_id not in comm_model.metabolites
                ):
                    new_met = Metabolite(m_id, met.name, ext_comp_id)
                    new_met.metadata = met.metadata.copy()
                    comm_model.add_metabolite(new_met)
                    ext_mets.append(new_met.id)

            # add genes
            for g_id, gene in model.genes.items():
                new_id = rename(g_id)
                new_gene = Gene(new_id, gene.name)
                new_gene.metadata = gene.metadata.copy()
                comm_model.add_gene(new_gene)

            # add internal reactions
            for r_id, rxn in model.reactions.items():

                new_id = rename(r_id)

                if rxn.reaction_type == ReactionType.EXCHANGE and r_id.startswith(
                    "R_EX"
                ):
                    new_id = new_id + self.suffix

                    m_id = list(rxn.stoichiometry.keys())[0]
                    ext_m_id = biomass_id if r_id == model.biomass_reaction else m_id
                    int_m_id = rename(m_id)

                    new_stoichiometry = {
                        ext_m_id: 1,
                        int_m_id: -1,
                    }
                    new_rxn = CBReaction(
                        new_id,
                        reversible=True,
                        stoichiometry=new_stoichiometry,
                        reaction_type=ReactionType.EXCHANGE,
                    )
                else:
                    new_stoichiometry = {
                        rename(m_id): coeff for m_id, coeff in rxn.stoichiometry.items()
                    }

                    if r_id == model.biomass_reaction:
                        new_stoichiometry[biomass_id] = 1

                    if rxn.gpr is None:
                        new_gpr = None
                    else:
                        new_gpr = GPRAssociation()
                        new_gpr.metadata = rxn.gpr.metadata.copy()

                        for protein in rxn.gpr.proteins:
                            new_protein = Protein()
                            new_protein.genes = [rename(g_id) for g_id in protein.genes]
                            new_protein.metadata = protein.metadata.copy()
                            new_gpr.proteins.append(new_protein)

                    new_rxn = CBReaction(
                        new_id,
                        name=rxn.name,
                        reversible=rxn.reversible,
                        stoichiometry=new_stoichiometry,
                        reaction_type=rxn.reaction_type,
                        lb=rxn.lb,
                        ub=rxn.ub,
                        gpr_association=new_gpr,
                    )

                comm_model.add_reaction(new_rxn)
                new_rxn.metadata = rxn.metadata.copy()
                self.reaction_map[(org_id, r_id)] = new_id

        # Add exchange reactions

        for m_id in ext_mets:
            r_id = f"R_EX_{m_id[2:]}" if m_id.startswith("M_") else f"R_EX_{m_id}"
            rxn = CBReaction(
                r_id,
                reversible=True,
                stoichiometry={m_id: -1},
                reaction_type=ReactionType.EXCHANGE,
            )
            comm_model.add_reaction(rxn)

        return comm_model



"""Create lists with variables for communities and models."""
from math import copysign


def get_binary(community):
    """Binary variables for community."""
    return [f"y_{org_id}" for org_id in community.organisms.keys()]


def get_biomass(community):
    """Biomass reactions for communities."""
    return [community.merged_model.biomass_reaction] + [
        community.reaction_map[org_id, org_model.biomass_reaction]
        for org_id, org_model in community.organisms.items()
    ]


def get_exchange_reactions(model):
    """Exchange reactions for model."""
    return [r_id for r_id in model.reactions.keys() if r_id.startswith("R_EX_")]


def get_consuming_metabolite_reactions(community, organism):
    """Get consuming reactions for organism in community."""
    reactions = []
    metabolite_lookup = {}
    metabolites = organism.get_external_metabolites(from_reactions=True)
    for metabolite in metabolites:
        rxns = organism.get_metabolite_consumers(metabolite)
        reactions = reactions + [
            community.reaction_map[organism.id, rxn]
            for rxn in rxns
            if not rxn.startswith("R_EX_") and not rxn.startswith("R_sink_")
        ]
        metabolite_lookup.update(
            {
                rxn: {
                    "metabolite": metabolite,
                    "direction": copysign(
                        1,
                        community.merged_model.reactions[rxn].stoichiometry[metabolite],
                    ),
                }
                for rxn in rxns
            }
        )

    return reactions, metabolite_lookup


def get_producing_metabolite_reactions(community, organism):
    """Get producing reactions for organism in community."""
    reactions = []
    producer_lookup = {}
    metabolites = organism.get_external_metabolites(from_reactions=True)
    for metabolite in metabolites:
        rxns = organism.get_metabolite_producers(metabolite)
        reactions = reactions + [
            community.reaction_map[organism.id, rxn]
            for rxn in rxns
            if not rxn.startswith("R_EX_") and not rxn.startswith("R_sink_")
        ]
        producer_lookup.update(
            {
                rxn: {
                    "metabolite": metabolite,
                    "direction": copysign(
                        1,
                        community.merged_model.reactions[rxn].stoichiometry[metabolite],
                    ),
                }
                for rxn in rxns
            }
        )

    return reactions, producer_lookup
